import std/[json, jsonutils, options, strutils, times]
import timeutils, uuids

type
  MetaSource* = enum msTagInfo = "tag info", msFileLocation = "file location"

  BaseModel* = object of RootObj
    id*: UUID
    name*: string

  Album* = object of BaseModel
    imageUri*: Option[string]
    trackTotal*: int
    year*: Option[int]

  Artist* = object of BaseModel
    imageUri*: Option[string]

  Bookmark* = object of BaseModel
    playlistId*: UUID
    mediaFileId*: UUID
    playIdx*: int
    playTimeMs*: int
    userCreated*: bool
    createdAt*: DateTime
    lastUsed*: DateTime

  MediaFile* = object of BaseModel
    comment*: Option[string]
    dateAdded*: DateTime
    discNumber*: Option[string]
    fileHash*: string
    filePath*: string
    imageUri*: Option[string]
    lastPlayed*: Option[DateTime]
    metaInfoSource*: MetaSource
    playCount*: int
    presentLocally*: bool
    trackNumber*: Option[int]

  Playlist* = object of BaseModel
    userCreated*: bool
    mediaFileCount*: int
    copiedFromId: Option[UUID]
    createdAt*: DateTime
    lastUsed*: DateTime

  Tag* = object
    name*: string
    description*: Option[string]

proc `$`*(mf: MediaFile): string =
  if mf.trackNumber.isSome: align($mf.trackNumber.get, 2) & " - " & mf.name
  else: mf.name

proc `$`*(a: Album): string =
  result = a.name
  if a.year.isSome: result &= " (" & $a.year & ")"

proc `$`*(m: BaseModel): string = m.name

proc `$`*(t: Tag): string =
  result = t.name
  if t.description.isSome: result &= t.description.get

proc toJsonHook(uuid: UUID): JsonNode = %($uuid)
proc fromJsonHook(u: var UUID, node: JsonNode) =
  u = parseUUID(node.getStr)

proc toJsonHook(dt: DateTime): JsonNode = %(dt.formatIso8601)
proc fromJsonHook(dt: var DateTime, node: JsonNode) =
  dt = parseIso8601(node.getStr)

proc `%`*(a: Album): JsonNode = toJson(a)
proc `%`*(a: Artist): JsonNode = toJson(a)
proc `%`*(b: Bookmark): JsonNode = toJson(b)
proc `%`*(mf: MediaFile): JsonNode =
  result = %*{
    "id": $mf.id,
    "name":  mf.name,
    "dateAdded": mf.dateAdded.formatIso8601,
    "fileHash": mf.fileHash,
    "filePath": mf.filePath,
    "metaInfoSource": $mf.metaInfoSource,
    "playCount": mf.playCount,
    "presentLocally": mf.presentLocally,
    "trackNumber": mf.trackNumber
  }

  if mf.comment.isSome: result["comment"] = %mf.comment.get
  if mf.discNumber.isSome: result["discNumber"] = %mf.discNumber.get
  if mf.imageUri.isSome: result["imageUri"] = %mf.imageUri.get
  if mf.lastPlayed.isSome: result["lastPlayed"] =
    %(mf.lastPlayed.get.formatIso8601)

proc `%`*(p: Playlist): JsonNode = toJson(p)
proc `%`*(t: Tag): JsonNode = toJson(t)

proc parseAlbum*(n: JsonNode): Album = result.fromJson(n)
proc parseArtist*(n: JsonNode): Artist = result.fromJson(n)
proc parseBookmark*(n: JsonNode): Bookmark = result.fromJson(n)

proc parseMediaFile*(n: JsonNode): MediaFile =
  MediaFile(
    id: n.parseUUID("id"),
    name: n.getOrFail("name").getStr,
    dateAdded: n.getOrFail("dateAdded").getStr.parseIso8601,
    discNumber:
      if n.contains("discNumber"): some(n["discNumber"].getStr)
      else: none[string](),
    fileHash: n.getOrFail("fileHash").getStr,
    filePath: n.getOrFail("filePath").getStr,
    imageUri:
      if n.contains("imageUrl"): some(n["imageUrl"].getStr)
      else: none[string](),
    lastPlayed:
      if n.contains("lastPlayed"): some(n["lastPlayed"].getStr.parseIso8601)
      else: none[DateTime](),
    metaInfoSource:
        parseEnum[MetaSource](n.getOrFail("metaInfoSource").getStr),
    playCount: n.getOrFail("playCount").getInt,
    presentLocally: n.getOrFail("presentLocally").getBool,
    trackNumber:
      if n.contains("trackNumber"): some(n["trackNumber"].getInt)
      else: none[int]())


proc parsePlaylist*(n: JsonNode): Playlist = result.fromJson(n)
proc parseTag*(n: JsonNode): Tag = result.fromJson(n)
